/**@fileoverview 
	SMIL Media element implementation.
*/

/**
 * Common code for SMIL media (and other) elements.
 * Hooks into the DOMElement through smil_playable_object attribute.
 * @constructor
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilPlayable(elmnt)
{
	this.elmnt = elmnt;
	this.playable_elmnt = null;
	if (elmnt) {
		elmnt.smil_playable_object = this;
	}
}

/**
 * Return HTML &lt;div&gt; that should receive output for this element.
 */
SmilPlayable.prototype.get_region = function ()
{
	var region_name = this.elmnt.getAttribute("region");
	var region;
	if (region_name !== null) {
		region = document.getElementById(region_name);
	}
	if (region === null) {
		this.elmnt.ownerDocument.logger.log("jsambulant: media items must have region attributes, for now");
		return;
	}
	if (region.get_region === undefined) {
		this.elmnt.ownerDocument.logger.log("jsambulant: media items region attribute must refer to region");
		return;
	}
	return region.get_region();
};

/**
 * Start playback of the media element.
 * Default (dummy) method which does nothing and calls the completion callback right away.
 * @param {function} playdone_callback Called upon completion of media playback.
 */
SmilPlayable.prototype.smilPlayable_start = function (playdone_callback)
{
	this.elmnt.ownerDocument.logger.log("jsambulant: smilPlayable_start not implemented yet for this element type");
	if (playdone_callback) {
		playdone_callback(this.elmnt);
	}
};

/**
 * Stop playback of media element.
 */
SmilPlayable.prototype.smilPlayable_stop = function smilPlayable_stop()
{
	if (this.playable_elmnt === null) {
		this.elmnt.ownerDocument.logger.log("jsambulant: smilPlayable_stop called while nothing is playing");
		return;
	}
	this.playable_elmnt.parentNode.removeChild(this.playable_elmnt);
	this.playable_elmnt = null;
};

/*global SmilPlayableRef*/
/**
 * Implement &lt;smil:ref&gt; element.
 * Currently a dummy placeholder.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
SmilPlayableRef = SmilPlayable;

/*global SmilPlayableAnimation*/
/**
 * Implement &lt;smil:animation&gt; element.
 * Currently a dummy placeholder.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
SmilPlayableAnimation = SmilPlayable;

/*global SmilPlayableTextstream*/
/**
 * Implement &lt;smil:textstream&gt; element.
 * Currently a dummy placeholder.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
SmilPlayableTextstream = SmilPlayable;

/**
 * Implement &lt;smil:audio&gt; element.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilPlayableAudio(elmnt)
{
	SmilPlayable.call(this, elmnt);
}

SmilPlayableAudio.prototype = new SmilPlayable();

/**
 * Start playback of the audio element.
 * Currently assumes that the QuickTime plugin is used (either IE or Firefox/other),
 * and no checks are made for it actually being available.
 * Arrangements are made for the completion callback being made on media end.
 * @param {function} playdone_callback Called upon completion of media playback.
 */
SmilPlayableAudio.prototype.smilPlayable_start = function (playdone_callback)
{
	var region = this.get_region();
	var src = this.elmnt.getAttribute("src");
	if (document.all) {
		// First create the IE element
		this.playable_elmnt = region.ownerDocument.createElement("object");
		this.playable_elmnt.setAttribute("CLASSID", "clsid:02BF25D5-8C17-4B23-BC80-D3488ABDDC6B");
		this.playable_elmnt.setAttribute("width", region.style.width);
		this.playable_elmnt.setAttribute("height", region.style.height);
		this.playable_elmnt.setAttribute("CODEBASE", "http://www.apple.com/qtactivex/qtplugin.cab");
		
		var param = region.ownerDocument.createElement("param");
		param.setAttribute("name", "src");
		param.setAttribute("value", src);
		this.playable_elmnt.appendChild(param);
		
		param = region.ownerDocument.createElement("param");
		param.setAttribute("name", "qtsrc");
		param.setAttribute("value", src);
		this.playable_elmnt.appendChild(param);
		
		param = region.ownerDocument.createElement("param");
		param.setAttribute("name", "controller");
		param.setAttribute("value", "false");
		this.playable_elmnt.appendChild(param);

		param = region.ownerDocument.createElement("param");
		param.setAttribute("name", "hidden");
		param.setAttribute("value", "true");
		this.playable_elmnt.appendChild(param);

		param = region.ownerDocument.createElement("param");
		param.setAttribute("name", "postdomevents");
		param.setAttribute("value", "false");
		this.playable_elmnt.appendChild(param);
	} else {
		this.playable_elmnt = region.ownerDocument.createElement("embed");
		this.elmnt.ownerDocument.logger.log("jsambulant: playing audio: " + src);
		this.playable_elmnt.setAttribute("src", src);
		this.playable_elmnt.setAttribute("width", region.style.width);
		this.playable_elmnt.setAttribute("height", region.style.height);
		this.playable_elmnt.setAttribute("controller", false);
		this.playable_elmnt.setAttribute("hidden", true);
		this.playable_elmnt.setAttribute("postdomevents", true);
	}
	region.appendChild(this.playable_elmnt);
	var that = this;
	if (playdone_callback) {
		if (this.playable_elmnt.addEventListener) { // Most browsers
			this.playable_elmnt.addEventListener("qt_ended", 
				function() { playdone_callback(that.elmnt); }, false);
		} else {	// IE
			this.playable_elmnt.attachEvent("onqt_ended",
				function() { playdone_callback(that.elmnt); });
		}
	}
};

/**
 * Implement &lt;smil:img&gt; element.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilPlayableImg(elmnt)
{
	SmilPlayable.call(this, elmnt);
}

SmilPlayableImg.prototype = new SmilPlayable();

/**
 * Start playback of the img element.
 * Inserts an HTML image element into the destination region, and calls
 * the callback immedeately.
 * @param {function} playdone_callback Called upon completion of media playback.
 */

SmilPlayableImg.prototype.smilPlayable_start = function (playdone_callback)
{
	var region = this.get_region();
	var src = this.elmnt.getAttribute("src");
	this.playable_elmnt = region.ownerDocument.createElement("img");
	this.elmnt.ownerDocument.logger.log("jsambulant: playing img: " + src);
	this.playable_elmnt.setAttribute("src", src);
	region.appendChild(this.playable_elmnt);
	if (playdone_callback) {
		playdone_callback(this.elmnt);
	}
};

/**
 * Implement &lt;smil:text&gt; element.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilPlayableText(elmnt)
{
	SmilPlayable.call(this, elmnt);
}

SmilPlayableText.prototype = new SmilPlayable();

/**
 * Start playback of the text element.
 * Inserts a DOMTextNode into the destination region, and calls
 * the callback immedeately.
 * @param {function} playdone_callback Called upon completion of media playback.
 */

SmilPlayableText.prototype.smilPlayable_start = function (playdone_callback)
{
	var region = this.get_region();
	var src = this.elmnt.getAttribute("src");
	var req = new XMLHttpRequest();
	req.open('GET', src, false);
	req.send(null);
	var text;
	if (req.status == 200 || req.status == 0) {
		text = req.responseText;
	} else {
		text = "error: code " + req.status;
	}	
	this.playable_elmnt = region.ownerDocument.createTextNode(text);
	this.elmnt.ownerDocument.logger.log("jsambulant: playing text: " + text);
	region.appendChild(this.playable_elmnt);
	if (playdone_callback) {
		playdone_callback(this.elmnt);
	}
};

/**
 * Implement &lt;smil:video&gt; element.
 * @constructor
 * @extends SmilPlayable
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilPlayableVideo(elmnt)
{
	SmilPlayable.call(this, elmnt);
}

SmilPlayableVideo.prototype = new SmilPlayable();

/**
 * Start playback of the video element.
 * Currently assumes that the QuickTime plugin is used (either IE or Firefox/other),
 * and no checks are made for it actually being available.
 * Arrangements are made for the completion callback being made on media end.
 * @param {function} playdone_callback Called upon completion of media playback.
 */
SmilPlayableVideo.prototype.smilPlayable_start = function (playdone_callback)
{
	var region = this.get_region();
	var src = this.elmnt.getAttribute("src");
	this.playable_elmnt = region.ownerDocument.createElement("embed");
	this.elmnt.ownerDocument.logger.log("jsambulant: playing video: " + src);
	this.playable_elmnt.setAttribute("src", src);
	this.playable_elmnt.setAttribute("width", region.style.width);
	this.playable_elmnt.setAttribute("height", region.style.height);
	this.playable_elmnt.setAttribute("controller", false);
	this.playable_elmnt.setAttribute("hidden", false);
	this.playable_elmnt.setAttribute("postdomevents", true);
	region.appendChild(this.playable_elmnt);
	var that = this;
	if (playdone_callback) {
		this.playable_elmnt.addEventListener("qt_ended", 
				function() { playdone_callback(that.elmnt); }, false);
	}
};
