/**@fileoverview 
	SMIL scheduler implementation.
*/

/*global window*/
/**
 * Common code for SMIL interior (par/seq) and media elements.
 * Hooks into the DOMElement through smil_sched_object attribute.
 * For interior elements, also hooks there are smil_playable_object (this
 * is a bit of a hack), so all executable smil elements have the smil_playable_object
 * API.
 * @constructor
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilSched(elmnt)
{
	this.elmnt = elmnt;
	this.playdone_callback = null;
	if (elmnt && !elmnt.smil_playable_object) {
		// Only do this for par/seq/etc, not media
		elmnt.smil_playable_object = this;
	}
	if (elmnt) {
		elmnt.smil_sched_object = this;
	}
}

/**
 * Returns the value of a DOM attribute on the ownned element, as seconds.
 * Parses a (subset of) SMIL time syntax.
 * @param {string} attrname The attribute.
 * @returns {number} Time in seconds.
 */
SmilSched.prototype.get_time = function (attrname)
{
	var value = this.elmnt.getAttribute(attrname);
	if (value === null) {
		return value;
	}
	var parse_clockval = /^([0-9]+(?:\.[0-9]+)?)(h|min|s|ms)?$/;
	var result = parse_clockval.exec(value);
	var sec = Number(result[1]);
	if (result[2] == "h") {
		sec = sec*3600;
	} else if (result[1] == "min") {
		sec = sec*60;
	} else if (result[1] == "ms") {
		sec = sec/1000;
	}
	//this.elmnt.ownerDocument.logger.log("jsambulant: get_time parse " + result[1] + "," + result[2] + "=" + sec);
	return sec;
};

/**
 * Start a given child element.
 * Caters for dur and begin attributes, and makes sure child_stopped() is called upon
 * either child completion or duration specified in the dur attribute.
 * @param {DOMElement} elmnt The child to start.
 */
SmilSched.prototype.start_child = function(elmnt)
{
	if (elmnt.smil_playable_object === undefined ||
		elmnt.smil_sched_object === undefined) {
		// Not a smil node. Ignore it.
		this.child_stopped(elmnt);
		return;
	}
	var child_dur = elmnt.smil_sched_object.get_time("dur");
	var child_begin = elmnt.smil_sched_object.get_time("begin");
	var _func1;
	var that = this;
	
	if (child_dur === null) {
		/** @ignore */
		_func1 = function() {
			elmnt.smil_playable_object.smilPlayable_start(function(elmnt) {that.child_stopped(elmnt); });
		};
	} else {
		child_dur = child_dur*1000;
		/** @ignore */
		_func1 = function() {
			elmnt.smil_playable_object.smilPlayable_start(undefined);
			window.setTimeout(function() { that.child_stopped(elmnt); }, child_dur);
		};
	}
	
	if (child_begin === null) {
		_func1();
	} else {
		window.setTimeout(_func1, child_begin*1000);
	}
};

/**
 * Stop playback of a child.
 * @param {DOMElement} elmnt The child to stop.
 */
SmilSched.prototype.stop_child = function(elmnt)
{
	if (elmnt.smil_playable_object === undefined) {
		return;
	}
	elmnt.smil_playable_object.smilPlayable_stop();
};

/**
 * Implementation of &lt;smil:par&gt; element.
 * @constructor
 * @extends SmilSched
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilSchedPar(elmnt)
{
	SmilSched.call(this, elmnt);
	this.children_active = 0;
}

SmilSchedPar.prototype = new SmilSched();

/*
 * Start playback of a SMIL par.
 * Playback for each of its children is started through start_child().
 */
SmilSchedPar.prototype.smilPlayable_start = function(playdone_callback)
{
	var childNodes = this.elmnt.childNodes;
	this.playdone_callback = playdone_callback;
	this.children_active = childNodes.length;
	if (this.children_active == 0) {
		if (this.playdone_callback) {
			playdone_callback(this.elmnt);
		}
		return;
	}
	for (var i=0; i<childNodes.length; i++) {
		this.start_child(childNodes[i]);
	}
};

/**
 * Stop playback of a par.
 * All children are stopped, through stop_child().
 */
SmilSchedPar.prototype.smilPlayable_stop = function()
{
	var childNodes = this.elmnt.childNodes;
	for (var i=0; i<childNodes.length; i++) {
		this.stop_child(childNodes[i]);
	}
};

/**
 * Called when a par child terminates.
 * Keeps track of the number of active children, and when this reaches zero calls
 * the playdone_callback for the current par itself.
 */
SmilSchedPar.prototype.child_stopped = function(elmnt)
{
	this.children_active--;
	if (this.children_active <= 0) {
		if (this.playdone_callback) {
			this.playdone_callback(this.elmnt);
		}
		this.playdone_callback = undefined;
	}
};			

/**
 * Implementation of &lt;smil:seq&gt; element.
 * @constructor
 * @extends SmilSched
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilSchedSeq(elmnt)
{
	SmilSched.call(this, elmnt);
}

SmilSchedSeq.prototype = new SmilSched();

/**
 * Start playback of a SMIL seq.
 * Playback of the first child starts, through start_child().
 */
SmilSchedSeq.prototype.smilPlayable_start = function(playdone_callback)
{
	this.curChild = this.elmnt.firstChild;
	if (this.curChild === null) {
		playdone_callback();
		return;
	}
	this.playdone_callback = playdone_callback;
	this.start_child(this.curChild);
};

/**
 * Stop playback of a seq.
 * The currently playing child, if any, is stopped.
 */
SmilSchedSeq.prototype.smilPlayable_stop = function()
{
	var curChild = this.curChild;
	this.curChild = null;
	if (curChild !== null) {
		this.stop_child(curChild);
	}
};

/**
 * Called when a seq child terminates.
 * If this is the last child in the seq the, the plaback_done callback for the seq itself
 * is issued. Otherwise, the currently playing child is stopped and the next one started.
 */
SmilSchedSeq.prototype.child_stopped = function(elmnt)
{
	if (!this.curChild) {
		this.elmnt.ownerDocument.logger.log("jsambulant: child_stopped for <seq> with no active child");
		return;
	}
	this.curChild = this.curChild.nextSibling;
	if (!this.curChild) {
		if (this.playdone_callback) {
			this.playdone_callback(this.elmnt);
		}
		return;
	}
	if (elmnt) {
		this.stop_child(elmnt);
	}
	this.start_child(this.curChild);
};	

/**
 * Scheduler portion of media and other non-interior SMIL nodes.
 * Provides get_time and other functionality through the base class.
 * @constructor
 * @extends SmilSched
 * @param {DOMElement} elmnt The DOM element to attach to.
 */
function SmilSchedMedia(elmnt)
{
	SmilSched.call(this, elmnt);
}
SmilSchedMedia.prototype = new SmilSched();
