/**@fileoverview 
	Implementation of SMIL Layout.
*/

/**
 * SMIL &lt;smil:layout&gt; element.
 * Creates toplevel layout &lt;div&gt;.
 * @constructor
 * @param {DOMElement} elmnt The DOM element.
 */
function SmilLayout(elmnt)
{
	this.elmnt = elmnt;
	// Create toplevel container (to be filled later)
	var region = document.createElement("div");
	region.style.position = "relative";
	region.style.width = 128;
	region.style.height = 128;
	region.style.top = 0;
	region.style.left = 0;
	this.region_elmnt = region;
	this.elmnt.appendChild(this.region_elmnt);
}

/**
 * Return HTML &lt;div&gt; that should receive output to this SMIL region.
 */
SmilLayout.prototype.get_region = function get_region()
{
	return this.region_elmnt;
};

/**
 * SMIL region/toplayout elements, common code. 
 * @constructor
 * @param {DOMElement} elmnt The DOM element.
*/
function SmilRegionCommon(elmnt)
{
	this.elmnt = elmnt;
	this.region_elmnt = null;
}

/**
 * Return HTML &lt;div&gt; that should receive output to this SMIL region.
 */
SmilRegionCommon.prototype.get_region = function get_region()
{
	return this.region_elmnt;
};


/**
 * Initialize width, height and other attributes common to region and topLayout.
 */
SmilRegionCommon.prototype.init_region = function init_region(elmnt)
{
	var width = elmnt.getAttribute("width");
	var height = elmnt.getAttribute("height");
	var backgroundColor = elmnt.getAttribute("backgroundColor");
	var zIndex = elmnt.getAttribute("zIndex");
	if (backgroundColor === null) {
		backgroundColor = elmnt.getAttribute("background-color");
	}
	if (zIndex === null) {
		zIndex = elmnt.getAttribute("z-index");
	}
	this.region_elmnt.style.display = "block";
	if (width !== null) {
		this.region_elmnt.style.width = width;
	}
	if (height !== null) {
		this.region_elmnt.style.height = height;
	}
	if (backgroundColor !== null) {
		this.region_elmnt.style.backgroundColor = backgroundColor;
	}
	if (zIndex !== null) {
		this.region_elmnt.style.zIndex = zIndex;
	}
//	/*DBG*/ this.region_elmnt.style.borderStyle = "dotted";
};

/**
 * &lt;smil:region&gt; element implementation.
 * Creates subregion and inserts it into toplevel layout or parent region.
 * @constructor
 * @extends SmilRegionCommon
 * @param {DOMElement} elmnt The DOM element.
*/
function SmilRegion(elmnt)
{
	var parent_region;
	if (elmnt.parentNode === null || elmnt.parentNode.get_region === null) {
		document.logger.log("jsambulant: parent of region should be a region, topLayout or layout");
		debugger;
	}
	parent_region = elmnt.parentNode.get_region();
	if (!parent_region) {
		document.logger.log("jsambulant: parent region not correctly initialized");
		debugger;
	}
	SmilRegionCommon.call(this, elmnt);
	this.region_elmnt = document.createElement("div");
	this.init_region(elmnt);
	this.region_elmnt.style.position = "absolute";
	var left = elmnt.getAttribute("left");
	var top = elmnt.getAttribute("top");
	var right = elmnt.getAttribute("right");
	var bottom = elmnt.getAttribute("bottom");
	if (left !== null) {
		this.region_elmnt.style.left = left;
	}
	if (top !== null) {
		this.region_elmnt.style.top = top;
	}
	if (right !== null) {
		this.region_elmnt.style.right = right;
	}
	if (bottom !== null) {
		this.region_elmnt.style.bottom = bottom;
	}
	parent_region.appendChild(this.region_elmnt);
}

SmilRegion.prototype = new SmilRegionCommon();

/**
 * &lt;smil:topLayout&gt; and &lt;smil:root-layout&gt; element implementation.
 * Modifies toplevel layout &lt;div&gt; to have correct width, height, etc.
 * @constructor
 * @extends SmilRegionCommon
 * @param {DOMElement} elmnt The DOM element.
*/
function SmilTopLayout(elmnt)
{
	var region_elmnt = elmnt.parentNode.get_region();
	if (!region_elmnt) {
		debugger;
	}
	SmilRegionCommon.call(this, elmnt);
	this.region_elmnt = region_elmnt;
	this.init_region(elmnt);
}

SmilTopLayout.prototype = new SmilRegionCommon();
